Set-StrictMode -Version Latest
#updated 20230807
#oagGameFields.ps1

[byte]$miVerPN = 8

function Confirm-FieldAllowsMultipleValues([parameter(mandatory)][ValidateSet('Categories','Companies','Features','Genres','Series','Tags','Developers','Publishers','Regions')][string]$FieldName)
{
# not compatible with Playnite 9+
	[bool]$bResult = $true

	switch ($miVerPN)
	{
		8 {
			$bResult = switch ($FieldName)
			{
				'Series' {$false}
				'Platform' {$false}
				'Age Rating' {$false}
				'Source' {$false}
				'Regions' {$false}
				default {$true}
			}
		}
		10 {
			$bResult = switch ($FieldName)
			{
				'Series' {$true}
				'Platform' {$true}
				'Age Rating' {$true}
				'Source' {$false}
				'Regions' {$true}
				default {$true}
			}
		}
	}

	return $bResult
}

function Get-DatabaseFieldGUID(
	[parameter(mandatory)][ValidateSet('Categories','Companies','Features','Genres','Series','Tags','Developers','Publishers','Regions','Platforms')][string]$FieldName,
	[parameter(mandatory)][string]$value,
	[switch]$Force
) {
#updated 20230807
	[string]$sResult = ''
	[string]$sActualFieldName = ''

	if ($FieldName -in @('Publishers','Developers'))
	{
		$sActualFieldName = 'Companies'
	} else {
		$sActualFieldName = $FieldName
	}

	try {
		$sResult = $($PlayniteApi.Database."$sActualFieldName" | Where-Object {$_.Name -eq $value}).Id
	} catch {}

	if (!$sResult)
	{
		if ($Force)
		{
			[object]$oItem = $PlayniteApi.Database."$sActualFieldName".Add($value)
			$sResult = $oItem.Id
		}
	}

	return $sResult
}

function Get-NameFieldIds(
	[parameter(mandatory)][ValidateSet('Categories','Features','Genres','Series','Tags','Developers','Publishers')][string]$FieldName,
	[bool]$bAssumeMultiValue = $false
) {
#updated 20221221
	[string]$sResult = ''
	[bool]$bMultiValueField = $true

	if ($bAssumeMultiValue)
	{
		$bMultiValueField = $true
	} else {
		$bMultiValueField = Confirm-FieldAllowsMultipleValues -FieldName $FieldName
	}

	$sResult = Get-FieldNameSingular -FieldName $FieldName

	if ($bMultiValueField)
	{
		$sResult = "$($sResult)Ids"
	} else {
		$sResult = "$($sResult)Id"
	}

	return $sResult
}

function Get-FieldNameSingular([parameter(mandatory)][ValidateSet('Categories','Features','Genres','Series','Tags','Developers','Publishers')][string]$FieldName)
{
#updated 20221221
	[string]$sResult = ''
	[bool]$bReturnLowerCase = $true

	$sResult = switch ($FieldName)
	{
		'Categories' {'Category'}
		'Series' {'Series'}
		default {$($FieldName.Substring(0,$FieldName.Length - 1))}
	}

	if ($bReturnLowerCase)
	{
		$sResult = $sResult.ToLower()
	}

	return $sResult
}

function Delete-FieldValueFromDatabase(
	[parameter(mandatory)][ValidateSet('Companies','Features','Genres','Tags')][string]$FieldName,
	[string]$value
) {
#updated 20230807
#required functions: Confirm-FieldAllowsMultipleValues , Get-DatabaseFieldGUID
#notes: can not remove single value fields

	[bool]$bResult = $false

	if ((Confirm-FieldAllowsMultipleValues -FieldName $FieldName) -ne $true)
	{
		return $bResult
	}

	if (![String]::IsNullorEmpty($value))
	{

		[object[]]$GamesWithValue = @()
		$GamesWithValue = @($PlayniteApi.Database.Games | Where-Object {$_."$FieldName".Name -eq $value})

		if (@($GamesWithValue).count -gt 0)
		{
			$__logger.Warn("unable to remove $FieldName '$value' from database as it is still in use")
			return $bResult
		}

		$idDbField = Get-DatabaseFieldGUID -FieldName $FieldName -value $value
		if ($idDbField)
		{
			$null = $PlayniteApi.Database."$FieldName".Remove($idDbField)
			$bResult = $true
			$__logger.Info("Removed $FieldName '$value' from database")
		}
	}

	return $bResult
}

function Get-MsgUpdatedGameField(
	[parameter(mandatory)][object]$game,
	[parameter(mandatory)][ValidateSet('Categories','Features','Genres','Series','Tags','Developers','Publishers')][string]$FieldName,
	[parameter(mandatory)][string]$value,
	[parameter(mandatory)][ValidateSet('Added','Removed','Renamed','Moved')][string]$sAction,
	[string]$value2,
	[ValidateSet('Categories','Features','Genres','Series','Tags','Developers','Publishers')][string]$FieldName2
) {
#updated 20221221

	[string]$sResult = ''

	$sResult = switch ($sAction)
	{
		'Renamed' {"{0}{1} {2} {3} to '$value2' for {4} {5}"}
		'Moved' {}
		'Added' {"{0}{1} {2} {3} to {4} {5}"}
		'Removed' {"{0}{1} {2} {3} from {4} {5}"}
	}

	$sResult = $sResult -f "Updated game. ", $sAction, $(Get-FieldNameSingular($FieldName)), "'$value'", "'$game'", "($($game.Platform.name))"

	return $sResult
}

function Invoke-AddRemoveTagsUsingHashGameIdAndBool(
	[parameter(mandatory)][hashtable]$hashResults,
	[parameter(mandatory)][string]$sChildPathTemplate,
	[parameter(mandatory)][string]$sTagValue,
	[parameter(mandatory)][bool]$bTagMissing,
	[bool]$bAutoApplyOppositeAction = $true,
	[bool]$bSilent = $false
) {
#updated 20230727
# required input: $hashResults must use game.id, [bool]$true/$false

	[string]$sMsg = ''
	[int]$iCountExistingLinked = 0
	[int]$iCountUnlinked = 0
	[int]$iTotalItemsProcessed = 0
	[int]$iCountTagsAdded = 0
	[int]$iCountTagsRemoved = 0
	[bool]$bAddTag = $false
	[bool]$bRemoveTag = $false
	[string]$sItemType = ''

	$StopWatch = [system.diagnostics.stopwatch]::StartNew()

	# set variables
	if ($($sChildPathTemplate.EndsWith('\')))
	{
		$sItemType = 'folder'
	} else {
		$sItemType = 'file'
	}

	ForEach ($h in $hashResults.GetEnumerator())
	{

		$iTotalItemsProcessed++

		# reset
		$bLinkedItemExists = $false
		$bAddTag = $false
		$bRemoveTag = $false
		$game = $null

		$bLinkedItemExists = $($h.value)
		$game = $PlayniteApi.Database.Games | Where-Object { ($_.Id -eq $h.key) } | Select-Object -First 1

		if ($game)
		{

			# update count
			if ($bLinkedItemExists)
			{
				$iCountExistingLinked++
			} else {
				$iCountUnlinked++
			}

			# check actions needed
			if ($bTagMissing)
			{
				$bRemoveTag = $bLinkedItemExists
				if ($bAutoApplyOppositeAction)
				{
					$bAddTag = !$bRemoveTag
				}
			} else {
				$bAddTag = $bLinkedItemExists
				if ($bAutoApplyOppositeAction)
				{
					$bRemoveTag = !$bAddTag
				}
			}

			# apply actions
			if ($bAddTag)
			{
				if ((Add-FieldValueToGame -game $game -FieldName 'Tags' -value $sTagValue) -eq $true)
				{
					$iCountTagsAdded++
				}
			} elseif ($bRemoveTag -eq $true) {
				if ((Remove-FieldValueFromGame -game $game -FieldName 'Tags' -value $sTagValue) -eq $true)
				{
					$iCountTagsRemoved++
				}
			}
		} else {
			$__logger.Error("no game found in database with id '$($h.key)'")
		}
	}

	# display results
	$sMsg = "$iTotalItemsProcessed games processed"
	$sMsg = $sMsg + "`r`n`r`n$iCountExistingLinked games with a matching $sItemType"
	$sMsg = $sMsg + "`r`n$iCountUnlinked games without a matching $sItemType"
	$sMsg = $sMsg + "`r`n`r`n$iCountTagsAdded games had tags added`r`n$iCountTagsRemoved games had tags removed"
	if ($iCountTagsAdded -gt 0)
	{
		$sMsg = $sMsg + "`r`n`r`n(Tag applied = '$sTagValue')"
	}
	$sMsg = $sMsg + "`r`n`r`n(running time $($StopWatch.Elapsed))"
	if (!$bSilent)
	{
		$null = $PlayniteApi.Dialogs.ShowMessage($sMsg)
	}
	$__logger.Info($sMsg.Replace("`r`n",' '))
}

function Add-FieldValueToGame(
	[parameter(mandatory)][object]$game,
	[parameter(mandatory)][ValidateSet('Features','Genres','Series','Tags','Developers','Publishers')][string]$FieldName,
	[string]$value
) {
#updated 20230109

	[bool]$bResult = $false
	[bool]$bCorrectResult = $false
	[bool]$bEditedGame = $false
	[int]$iAction = -1

	$value = $value.TrimEnd()
	if ([String]::IsNullorEmpty($value))
	{
	} else {
		$iAction = Add-FieldValueToGameNoUpdate -game $game -FieldName $FieldName -value $value
		switch ($iAction)
		{
			0 {
				$bCorrectResult = $false
				$bEditedGame = $false
			}
			1 {
				$bCorrectResult = $true
				$bEditedGame = $false
			}
			2 {
				$bCorrectResult = $true
				$bEditedGame = $true
			}
		}
	}

	if (($bCorrectResult) -and ($bEditedGame))
	{
		$game.Modified = (Get-Date)
		$null = $PlayniteApi.Database.Games.Update($game)
		[string]$sMsg = Get-MsgUpdatedGameField -game $game -FieldName $FieldName -value $value -sAction 'Added'
		$__logger.Info($sMsg)
		$bResult = $true
	}

	return $bResult
}

function Rename-FieldValueForGame(
[parameter(mandatory)][object]$game,
	[parameter(mandatory)][ValidateSet('Categories','Features','Genres','Series','Tags','Developers','Publishers')][string]$FieldName,
	[parameter(mandatory)][string]$sValueBefore,
	[parameter(mandatory)][string]$sValueAfter
) {
#updated 20230109

	[bool]$bResult = $false
	[bool]$bCorrectResult = $false
	[bool]$bEditedGame = $true
	[bool]$bEditedGameAdd = $false
	[int]$iAction = -1

	if ($(Confirm-GameHasFieldValue -game $game -FieldName $FieldName -value $sValueBefore) -ne $true)
	{
		return $bResult
	}

	$sValueAfter = $sValueAfter.TrimEnd()

	$iAction = Add-FieldValueToGameNoUpdate -game $game -FieldName $FieldName -value $sValueAfter
	switch ($iAction)
	{
		0 {
			$bCorrectResult = $false
			$bEditedGame = $false
		}
		1 {
			$bCorrectResult = $true
			$bEditedGame = $false
		}
		2 {
			$bCorrectResult = $true
			$bEditedGame = $true
		}
	}

	$bEditedGameAdd = $bEditedGame

	if (!$bCorrectResult)
	{
		# failed to add new value
		return $bResult
	}


	if ($(Confirm-FieldAllowsMultipleValues -FieldName $FieldName) -eq $true)
	{

		$iAction = -1
		$iAction = Remove-FieldValueFromGameNoUpdate -game $game -FieldName $FieldName -value $sValueBefore
		switch ($iAction)
		{
			0 {
				$bCorrectResult = $false
				$bEditedGame = $false
			}
			1 {
				$bCorrectResult = $true
				$bEditedGame = $false
			}
			2 {
				$bCorrectResult = $true
				$bEditedGame = $true
			}
		}
	} else {
		#$bCorrectResult remains unchanged
	}

	if (($bCorrectResult) -or ($bEditedGameAdd -eq $true))
	{
		$game.Modified = (Get-Date)
		$null = $PlayniteApi.Database.Games.Update($game)
		if (($bCorrectResult -ne $true) -and ($bEditedGameAdd -eq $true))
		{
			# new value added but old value did not remove
			[string]$sMsg = Get-MsgUpdatedGameField -game $game -FieldName $FieldName -value $sValueAfter -sAction 'Added'
		} else {
			[string]$sMsg = Get-MsgUpdatedGameField -game $game -FieldName $FieldName -value $sValueBefore -sAction 'Renamed' -value2 $sValueAfter
		}
		$__logger.Info($sMsg)
		$bResult = $true
	}

	return $bResult
}

function Add-FieldValueToGameNoUpdate(
	[parameter(mandatory)][object]$game,
	[parameter(mandatory)][ValidateSet('Features','Genres','Series','Tags','Developers','Publishers')][string]$FieldName,
	[parameter(mandatory)][string]$value
) {
#updated 20230109

	[int]$iResult = -1
	[bool]$bCorrectResult = $false
	[bool]$bEditedGame = $false
	[bool]$bMultiValueField = $true
	[string]$sFieldIdName = ''

	$idDbField = Get-DatabaseFieldGUID -FieldName $FieldName -value $value -Force
	$bMultiValueField = Confirm-FieldAllowsMultipleValues -FieldName $FieldName
	$sFieldIdName = Get-NameFieldIds -FieldName $FieldName -bAssumeMultiValue $bMultiValueField

	if ($bMultiValueField)
	{
		if ($game."$sFieldIdName")
		{
			if ($game."$sFieldIdName" -contains $idDbField)
			{
				$bCorrectResult = $true
			} else {
				$null = $game."$sFieldIdName".Add($idDbField)
				$bEditedGame = $true
			}
		} else {
			$game."$sFieldIdName" = $idDbField
			$bEditedGame = $true
		}
		if ($bEditedGame)
		{
			# validate
			$bCorrectResult = ($game."$sFieldIdName" -contains $idDbField)
		}
	} else {
		if ($game."$sFieldIdName" -eq $idDbField)
		{
			$bCorrectResult = $true
		} else {
			$game."$sFieldIdName" = $idDbField
			$bEditedGame = $true

			# validate
			$bCorrectResult = ($game."$sFieldIdName" -eq $idDbField)
		}
	}

	if ($bCorrectResult)
	{
		if ($bEditedGame)
		{
			$iResult = 2
		} else {
			$iResult = 1
		}
	} else {
		$iResult = 0
		if ($bEditedGame)
		{
			$iResult = -1
			$__logger.Error("game edit but result fail $game")
		}
	}

	return $iResult
}

function Confirm-GameHasDeletedCategory([parameter(mandatory)][object]$game)
{
#updated 20221221

	[bool]$bResult = $false
	[string]$sConstant = 'DELETED'

	if ($game.Categories)
	{
		$bResult = $(Confirm-GameHasFieldValue -game $game -Value $sConstant -FieldName 'Categories')
	}

	return $bResult
}

function Confirm-GameHasFieldValue(
	[parameter(mandatory)][object]$game,
	[parameter(mandatory)][ValidateSet('Categories','Features','Genres','Tags','Developers','Publishers','Series')][string]$FieldName,
	[string]$value
) {
#updated 20230807

	[bool]$bResult = $false
	[string]$sFieldIdName = ''
	[bool]$bAllowsMultipleValues = $false

	if (!$value)
	{
		return $bResult
	}

	$bAllowsMultipleValues = Confirm-FieldAllowsMultipleValues -FieldName $FieldName

	if ($bAllowsMultipleValues)
	{
		$idDbField = Get-DatabaseFieldGUID -FieldName $FieldName -value $value
		if ($idDbField)
		{
			$sFieldIdName = Get-NameFieldIds -FieldName $FieldName
			$bResult = ($game."$sFieldIdName" -contains $idDbField)
		}
	} else {
		$bResult = ($game."$FieldName" -eq $value)
	}

	return $bResult
}

function Remove-FieldValueFromGame(
	[parameter(mandatory)][object]$game,
	[parameter(mandatory)][ValidateSet('Categories','Features','Genres','Tags','Developers','Publishers')][string]$FieldName,
	[string]$value
) {
#updated 20230109

	[bool]$bResult = $false
	[bool]$bCorrectResult = $true
	[bool]$bEditedGame = $false
	[int]$iAction = -1

	if ([String]::IsNullorEmpty($value))
	{
		$bCorrectResult = $true
		$bEditedGame = $false
	} else {
		$iAction = Remove-FieldValueFromGameNoUpdate -game $game -FieldName $FieldName -value $value

		switch ($iAction)
		{
			0 {
				$bCorrectResult = $false
				$bEditedGame = $false
			}
			1 {
				$bCorrectResult = $true
				$bEditedGame = $false
			}
			2 {
				$bCorrectResult = $true
				$bEditedGame = $true
			}
		}
	}

	if (($bCorrectResult) -and ($bEditedGame))
	{
		$game.Modified = (Get-Date)
		$null = $PlayniteApi.Database.Games.Update($game)
		[string]$sMsg = Get-MsgUpdatedGameField -game $game -FieldName $FieldName -value $value -sAction 'Removed'
		$__logger.Info($sMsg)
		$bResult = $true
	}

	return $bResult
}

function Remove-FieldValueFromGameNoUpdate(
	[parameter(mandatory)][object]$game,
	[parameter(mandatory)][ValidateSet('Categories','Features','Genres','Tags','Developers','Publishers')][string]$FieldName,
	[parameter(mandatory)][string]$value
) {
#updated 20230109

	[int]$iResult = -1
	[bool]$bCorrectResult = $true
	[bool]$bEditedGame = $false
	[string]$sFieldIdName = ''

	$sFieldIdName = Get-NameFieldIds -FieldName $FieldName -bAssumeMultiValue $true

	if ($game."$sFieldIdName")
	{

		$idDbField = Get-DatabaseFieldGUID -FieldName $FieldName -value $value
		if ($game."$sFieldIdName" -contains $idDbField)
		{
			$bCorrectResult = $false

			# remove then confirm removal
			$null = $game."$sFieldIdName".Remove($idDbField)
			$bEditedGame = $true

			if ($game."$sFieldIdName")
			{
				$bCorrectResult = ($game."$sFieldIdName" -NotContains $idDbField)
			}
		}
	}

	if ($bCorrectResult)
	{
		if ($bEditedGame)
		{
			$iResult = 2
		} else {
			$iResult = 1
		}
	} else {
		$iResult = 0
		if ($bEditedGame)
		{
			$iResult = -1
			$__logger.Error("game edit but result fail $game")
		}
	}

	return $iResult
}